#ifndef BEGIN_MAPPER
// FIRST INCLUSION
class mapper
{
	struct swapping { unsigned p1, p2, size8k; bool enabled; };
	swapping swaps[4];
	unsigned mapto[4];
public:
	bool debug;
	// These don't update automatically.
	void maprom(unsigned size4k, unsigned at, unsigned value);
	void exchange(unsigned swapid, unsigned size4k, unsigned at1, unsigned at2, bool state);
	void update();
public:
	void reset();
	unsigned vrom4kcount, rom8kcount;
public:
	mapper();
	virtual void initialization() = 0;
	virtual void write(fWORD ofs, fBYTE value) = 0;
};
#include <map>
extern map<unsigned, mapper *>mappers;

#define BEGIN_MAPPER(n, prog) static class mapper_##n : public mapper { public: mapper_##n() { mappers[n] = this; } virtual void initialization() { prog; if(debug)fprintf(stderr, "This is the state after mapper initialization\n"); update(); }
#define END_MAPPER(n)  }map##n;
#else
// SECOND INCLUSION
mapper::mapper() : debug(false)
{
	reset();
}
void mapper::reset()
{
	memset(swaps, 0, sizeof swaps);
	
	mapto[0] = 0;
	mapto[1] = 1;
	mapto[2] = rom8kcount-2;
	mapto[3] = rom8kcount-1;
}
void mapper::maprom(unsigned size8k, unsigned at, unsigned value)
{
	for(unsigned a=0; a<size8k; ++a)
		mapto[at + a] = value * size8k + a;
}
void mapper::exchange(unsigned swapid, unsigned size8k, unsigned at1, unsigned at2, bool state)
{
	swaps[swapid].p1 = at1;
	swaps[swapid].p2 = at2;
	swaps[swapid].size8k = size8k;
	swaps[swapid].enabled = state;
}

void mapper::update()
{
	unsigned a;
	if(debug)fprintf(stderr, "New map:");
	for(a=0; a<4; ++a)
	{
		if(debug)fprintf(stderr, " %u", mapto[a]);
		nes.rompage[a] = &nes.rom[mapto[a] * 0x2000];
	}
	if(debug)fprintf(stderr, "\n");
	bool swapped=false;
	for(a=0; a<4; ++a)
		if(swaps[a].enabled)
		{
			unsigned p1 = swaps[a].p1;
			unsigned p2 = swaps[a].p2;
			for(unsigned b=0; b<swaps[a].size8k; ++b, ++p1, ++p2)
			{
				if(debug){swapped=true;fprintf(stderr, "Swapping the mapped pages %u and %u\n", p1, p2);}
				BYTE *p = nes.rompage[p1];
				nes.rompage[p1] = nes.rompage[p2];
				nes.rompage[p2] = p;
			}
		}
	if(swapped)
	{
    	fprintf(stderr, "Result:");
    	for(a=0; a<4; ++a)fprintf(stderr, " %u", (nes.rompage[a]-&nes.rom[0])/0x2000);
    	fprintf(stderr, "\n");
	}
}
#endif
